from __future__ import generators

"""
httplib2

A caching http interface that supports ETags and gzip
to conserve bandwidth.

Requires Python 2.3 or later

Changelog:
2007-08-18, Rick: Modified so it's able to use a socks proxy if needed.

"""

__author__ = "Joe Gregorio (joe@bitworking.org)"
__copyright__ = "Copyright 2006, Joe Gregorio"
__contributors__ = ["Thomas Broyer (t.broyer@ltgt.net)",
                    "James Antill",
                    "Xavier Verges Farrero",
                    "Jonathan Feinberg",
                    "Blair Zajac",
                    "Sam Ruby",
                    "Louis Nyffenegger"]
__license__ = "MIT"
__version__ = "$Rev: 259 $"

import re
import sys

try:
    import hashlib

    md = hashlib.md5()
except ImportError:
    # for Python << 2.5
    import md5

    md = md5.new()
import email
import email.Utils
import email.Message
import StringIO
import gzip
import zlib
import httplib
import urlparse
import base64
import os
import copy
import calendar
import time
import random

try:
    import hashlib

    md = hashlib.md5()
except ImportError:
    # for Python << 2.5
    import md5

    md = md5.new()
import hmac
from gettext import gettext as _
import socket

try:
    import socks
except ImportError:
    socks = None

if sys.version_info >= (2, 3):
    from iri2uri import iri2uri
else:
    def iri2uri(uri):
        return uri

__all__ = ['Http', 'Response', 'ProxyInfo', 'HttpLib2Error',
           'RedirectMissingLocation', 'RedirectLimit', 'FailedToDecompressContent',
           'UnimplementedDigestAuthOptionError', 'UnimplementedHmacDigestAuthOptionError',
           'debuglevel']


# The httplib debug level, set to a non-zero value to get debug output
debuglevel = 0

# Python 2.3 support
if sys.version_info < (2, 4):
    def sorted(seq):
        seq.sort()
        return seq

# Python 2.3 support
def HTTPResponse__getheaders(self):
    """Return list of (header, value) tuples."""
    if self.msg is None:
        raise httplib.ResponseNotReady()
    return self.msg.items()

if not hasattr(httplib.HTTPResponse, 'getheaders'):
    httplib.HTTPResponse.getheaders = HTTPResponse__getheaders

# All exceptions raised here derive from HttpLib2Error
class HttpLib2Error(Exception): pass

# Some exceptions can be caught and optionally 
# be turned back into responses. 
class HttpLib2ErrorWithResponse(HttpLib2Error):
    def __init__(self, desc, response, content):
        self.response = response
        self.content = content
        HttpLib2Error.__init__(self, desc)


class RedirectMissingLocation(HttpLib2ErrorWithResponse): pass


class RedirectLimit(HttpLib2ErrorWithResponse): pass


class FailedToDecompressContent(HttpLib2ErrorWithResponse): pass


class UnimplementedDigestAuthOptionError(HttpLib2ErrorWithResponse): pass


class UnimplementedHmacDigestAuthOptionError(HttpLib2ErrorWithResponse): pass


class RelativeURIError(HttpLib2Error): pass


class ServerNotFoundError(HttpLib2Error): pass

# Open Items:
# -----------
# Proxy support

# Are we removing the cached content too soon on PUT (only delete on 200 Maybe?)

# Pluggable cache storage (supports storing the cache in
#   flat files by default. We need a plug-in architecture
#   that can support Berkeley DB and Squid)

# == Known Issues ==
# Does not handle a resource that uses conneg and Last-Modified but no ETag as a cache validator.
# Does not handle Cache-Control: max-stale
# Does not use Age: headers when calculating cache freshness.


# The number of redirections to follow before giving up.
# Note that only GET redirects are automatically followed.
# Will also honor 301 requests by saving that info and never
# requesting that URI again.
DEFAULT_MAX_REDIRECTS = 5

# Which headers are hop-by-hop headers by default
HOP_BY_HOP = ['connection', 'keep-alive', 'proxy-authenticate', 'proxy-authorization', 'te', 'trailers',
              'transfer-encoding', 'upgrade']

def _get_end2end_headers(response):
    hopbyhop = list(HOP_BY_HOP)
    hopbyhop.extend([x.strip() for x in response.get('connection', '').split(',')])
    return [header for header in response.keys() if header not in hopbyhop]

URI = re.compile(r"^(([^:/?#]+):)?(//([^/?#]*))?([^?#]*)(\?([^#]*))?(#(.*))?")

def parse_uri(uri):
    """Parses a URI using the regex given in Appendix B of RFC 3986.

        (scheme, authority, path, query, fragment) = parse_uri(uri)
    """
    groups = URI.match(uri).groups()
    return (groups[1], groups[3], groups[4], groups[6], groups[8])


def urlnorm(uri):
    (scheme, authority, path, query, fragment) = parse_uri(uri)
    if not scheme or not authority:
        raise RelativeURIError("Only absolute URIs are allowed. uri = %s" % uri)
    authority = authority.lower()
    scheme = scheme.lower()
    if not path:
        path = "/"
        # Could do syntax based normalization of the URI before
    # computing the digest. See Section 6.2.2 of Std 66.
    request_uri = query and "?".join([path, query]) or path
    scheme = scheme.lower()
    defrag_uri = scheme + "://" + authority + request_uri
    return scheme, authority, request_uri, defrag_uri


# Cache filename construction (original borrowed from Venus http://intertwingly.net/code/venus/)
re_url_scheme = re.compile(r'^\w+://')
re_slash = re.compile(r'[?/:|]+')

def safename(filename):
    """Return a filename suitable for the cache.

    Strips dangerous and common characters to create a filename we
    can use to store the cache in.
    """

    try:
        if re_url_scheme.match(filename):
            if isinstance(filename, str):
                filename = filename.decode('utf-8')
                filename = filename.encode('idna')
            else:
                filename = filename.encode('idna')
    except UnicodeError:
        pass
    if isinstance(filename, unicode):
        filename = filename.encode('utf-8')
    filemd5 = md5.new(filename).hexdigest()
    filename = re_url_scheme.sub("", filename)
    filename = re_slash.sub(",", filename)

    # limit length of filename
    if len(filename) > 200:
        filename = filename[:200]
    return ",".join((filename, filemd5))

NORMALIZE_SPACE = re.compile(r'(?:\r\n)?[ \t]+')

def _normalize_headers(headers):
    return dict([(key.lower(), NORMALIZE_SPACE.sub(value, ' ').strip())  for (key, value) in headers.iteritems()])


def _parse_cache_control(headers):
    retval = {}
    if headers.has_key('cache-control'):
        parts = headers['cache-control'].split(',')
        parts_with_args = [tuple([x.strip() for x in part.split("=")]) for part in parts if -1 != part.find("=")]
        parts_wo_args = [(name.strip(), 1) for name in parts if -1 == name.find("=")]
        retval = dict(parts_with_args + parts_wo_args)
    return retval

# Whether to use a strict mode to parse WWW-Authenticate headers
# Might lead to bad results in case of ill-formed header value,
# so disabled by default, falling back to relaxed parsing.
# Set to true to turn on, usefull for testing servers.
USE_WWW_AUTH_STRICT_PARSING = 0

# In regex below:
#    [^\0-\x1f\x7f-\xff()<>@,;:\\\"/[\]?={} \t]+             matches a "token" as defined by HTTP
#    "(?:[^\0-\x08\x0A-\x1f\x7f-\xff\\\"]|\\[\0-\x7f])*?"    matches a "quoted-string" as defined by HTTP, when LWS have already been replaced by a single space
# Actually, as an auth-param value can be either a token or a quoted-string, they are combined in a single pattern which matches both:
#    \"?((?<=\")(?:[^\0-\x1f\x7f-\xff\\\"]|\\[\0-\x7f])*?(?=\")|(?<!\")[^\0-\x08\x0A-\x1f\x7f-\xff()<>@,;:\\\"/[\]?={} \t]+(?!\"))\"?
WWW_AUTH_STRICT = re.compile(
    r"^(?:\s*(?:,\s*)?([^\0-\x1f\x7f-\xff()<>@,;:\\\"/[\]?={} \t]+)\s*=\s*\"?((?<=\")(?:[^\0-\x08\x0A-\x1f\x7f-\xff\\\"]|\\[\0-\x7f])*?(?=\")|(?<!\")[^\0-\x1f\x7f-\xff()<>@,;:\\\"/[\]?={} \t]+(?!\"))\"?)(.*)$")
WWW_AUTH_RELAXED = re.compile(
    r"^(?:\s*(?:,\s*)?([^ \t\r\n=]+)\s*=\s*\"?((?<=\")(?:[^\\\"]|\\.)*?(?=\")|(?<!\")[^ \t\r\n,]+(?!\"))\"?)(.*)$")
UNQUOTE_PAIRS = re.compile(r'\\(.)')

def _parse_www_authenticate(headers, headername='www-authenticate'):
    """Returns a dictionary of dictionaries, one dict
    per auth_scheme."""
    retval = {}
    if headers.has_key(headername):
        authenticate = headers[headername].strip()
        www_auth = USE_WWW_AUTH_STRICT_PARSING and WWW_AUTH_STRICT or WWW_AUTH_RELAXED
        while authenticate:
            # Break off the scheme at the beginning of the line
            if headername == 'authentication-info':
                (auth_scheme, the_rest) = ('digest', authenticate)
            else:
                (auth_scheme, the_rest) = authenticate.split(" ", 1)
                # Now loop over all the key value pairs that come after the scheme,
            # being careful not to roll into the next scheme
            match = www_auth.search(the_rest)
            auth_params = {}
            while match:
                if match and len(match.groups()) == 3:
                    (key, value, the_rest) = match.groups()
                    auth_params[key.lower()] = UNQUOTE_PAIRS.sub(r'\1',
                                                                 value) # '\\'.join([x.replace('\\', '') for x in value.split('\\\\')])
                match = www_auth.search(the_rest)
            retval[auth_scheme.lower()] = auth_params
            authenticate = the_rest.strip()
    return retval


def _entry_disposition(response_headers, request_headers):
    """Determine freshness from the Date, Expires and Cache-Control headers.

    We don't handle the following:

    1. Cache-Control: max-stale
    2. Age: headers are not used in the calculations.

    Not that this algorithm is simpler than you might think
    because we are operating as a private (non-shared) cache.
    This lets us ignore 's-maxage'. We can also ignore
    'proxy-invalidate' since we aren't a proxy.
    We will never return a stale document as
    fresh as a design decision, and thus the non-implementation 
    of 'max-stale'. This also lets us safely ignore 'must-revalidate' 
    since we operate as if every server has sent 'must-revalidate'.
    Since we are private we get to ignore both 'public' and
    'private' parameters. We also ignore 'no-transform' since
    we don't do any transformations.
    The 'no-store' parameter is handled at a higher level.
    So the only Cache-Control parameters we look at are:

    no-cache
    only-if-cached
    max-age
    min-fresh
    """

    retval = "STALE"
    cc = _parse_cache_control(request_headers)
    cc_response = _parse_cache_control(response_headers)

    if request_headers.has_key('pragma') and request_headers['pragma'].lower().find('no-cache') != -1:
        retval = "TRANSPARENT"
        if 'cache-control' not in request_headers:
            request_headers['cache-control'] = 'no-cache'
    elif cc.has_key('no-cache'):
        retval = "TRANSPARENT"
    elif cc_response.has_key('no-cache'):
        retval = "STALE"
    elif cc.has_key('only-if-cached'):
        retval = "FRESH"
    elif response_headers.has_key('date'):
        date = calendar.timegm(email.Utils.parsedate_tz(response_headers['date']))
        now = time.time()
        current_age = max(0, now - date)
        if cc_response.has_key('max-age'):
            try:
                freshness_lifetime = int(cc_response['max-age'])
            except ValueError:
                freshness_lifetime = 0
        elif response_headers.has_key('expires'):
            expires = email.Utils.parsedate_tz(response_headers['expires'])
            if None == expires:
                freshness_lifetime = 0
            else:
                freshness_lifetime = max(0, calendar.timegm(expires) - date)
        else:
            freshness_lifetime = 0
        if cc.has_key('max-age'):
            try:
                freshness_lifetime = int(cc['max-age'])
            except ValueError:
                freshness_lifetime = 0
        if cc.has_key('min-fresh'):
            try:
                min_fresh = int(cc['min-fresh'])
            except ValueError:
                min_fresh = 0
            current_age += min_fresh
        if freshness_lifetime > current_age:
            retval = "FRESH"
    return retval


def _decompressContent(response, new_content):
    content = new_content
    try:
        encoding = response.get('content-encoding', None)
        if encoding in ['gzip', 'deflate']:
            if encoding == 'gzip':
                content = gzip.GzipFile(fileobj=StringIO.StringIO(new_content)).read()
            if encoding == 'deflate':
                content = zlib.decompress(content)
            response['content-length'] = str(len(content))
            del response['content-encoding']
    except IOError:
        content = ""
        raise FailedToDecompressContent(
            _("Content purported to be compressed with %s but failed to decompress.") % response.get('content-encoding')
            , response, content)
    return content


def _updateCache(request_headers, response_headers, content, cache, cachekey):
    if cachekey:
        cc = _parse_cache_control(request_headers)
        cc_response = _parse_cache_control(response_headers)
        if cc.has_key('no-store') or cc_response.has_key('no-store'):
            cache.delete(cachekey)
        else:
            info = email.Message.Message()
            for key, value in response_headers.iteritems():
                if key not in ['status', 'content-encoding', 'transfer-encoding']:
                    info[key] = value

            status = response_headers.status
            if status == 304:
                status = 200

            status_header = 'status: %d\r\n' % response_headers.status

            header_str = info.as_string()

            header_str = re.sub("\r(?!\n)|(?<!\r)\n", "\r\n", header_str)
            text = "".join([status_header, header_str, content])

            cache.set(cachekey, text)


def _cnonce():
    dig = md5.new("%s:%s" % (time.ctime(), ["0123456789"[random.randrange(0, 9)] for i in range(20)])).hexdigest()
    return dig[:16]


def _wsse_username_token(cnonce, iso_now, password):
    return base64.encodestring(sha.new("%s%s%s" % (cnonce, iso_now, password)).digest()).strip()


# For credentials we need two things, first 
# a pool of credential to try (not necesarily tied to BAsic, Digest, etc.)
# Then we also need a list of URIs that have already demanded authentication
# That list is tricky since sub-URIs can take the same auth, or the 
# auth scheme may change as you descend the tree.
# So we also need each Auth instance to be able to tell us
# how close to the 'top' it is.

class Authentication(object):
    def __init__(self, credentials, host, request_uri, headers, response, content, http):
        (scheme, authority, path, query, fragment) = parse_uri(request_uri)
        self.path = path
        self.host = host
        self.credentials = credentials
        self.http = http
        self.response = response
        self.headers = headers
        self.content = content

    def depth(self, request_uri):
        (scheme, authority, path, query, fragment) = parse_uri(request_uri)
        return request_uri[len(self.path):].count("/")

    def inscope(self, host, request_uri):
        # XXX Should we normalize the request_uri?
        (scheme, authority, path, query, fragment) = parse_uri(request_uri)
        return (host == self.host) and path.startswith(self.path)

    def request(self, method, request_uri, headers, content):
        """Modify the request headers to add the appropriate
        Authorization header. Over-rise this in sub-classes."""
        pass

    def response(self, response, content):
        """Gives us a chance to update with new nonces
        or such returned from the last authorized response.
        Over-rise this in sub-classes if necessary.

        Return TRUE is the request is to be retried, for
        example Digest may return stale=true.
        """
        return False


class BasicAuthentication(Authentication):
    def __init__(self, credentials, host, request_uri, headers, response, content, http):
        Authentication.__init__(self, credentials, host, request_uri, headers, response, content, http)

    def request(self, method, request_uri, headers, content):
        """Modify the request headers to add the appropriate
        Authorization header."""
        headers['authorization'] = 'Basic ' + base64.encodestring("%s:%s" % self.credentials).strip()


class DigestAuthentication(Authentication):
    """Only do qop='auth' and MD5, since that
    is all Apache currently implements"""

    def __init__(self, credentials, host, request_uri, headers, response, content, http):
        Authentication.__init__(self, credentials, host, request_uri, headers, response, content, http)
        challenge = _parse_www_authenticate(response)
        self.challenge = challenge['digest']
        qop = self.challenge.get('qop')
        self.challenge['qop'] = ('auth' in [x.strip() for x in qop.split()]) and 'auth' or None
        if self.challenge['qop'] is None:
            raise UnimplementedDigestAuthOptionError(_("Unsupported value for qop: %s." % qop))
        self.challenge['algorithm'] = self.challenge.get('algorithm', 'MD5')
        if self.challenge['algorithm'] != 'MD5':
            raise UnimplementedDigestAuthOptionError(
                _("Unsupported value for algorithm: %s." % self.challenge['algorithm']))
        self.A1 = "".join([self.credentials[0], ":", self.challenge['realm'], ":", self.credentials[1]])
        self.challenge['nc'] = 1

    def request(self, method, request_uri, headers, content, cnonce=None):
        """Modify the request headers"""
        H = lambda x: md5.new(x).hexdigest()
        KD = lambda s, d: H("%s:%s" % (s, d))
        A2 = "".join([method, ":", request_uri])
        self.challenge['cnonce'] = cnonce or _cnonce()
        request_digest = '"%s"' % KD(H(self.A1), "%s:%s:%s:%s:%s" % (self.challenge['nonce'],
                                                                     '%08x' % self.challenge['nc'],
                                                                     self.challenge['cnonce'],
                                                                     self.challenge['qop'], H(A2)
        ))
        headers[
        'Authorization'] = 'Digest username="%s", realm="%s", nonce="%s", uri="%s", algorithm=%s, response=%s, qop=%s, nc=%08x, cnonce="%s"' % (
        self.credentials[0],
        self.challenge['realm'],
        self.challenge['nonce'],
        request_uri,
        self.challenge['algorithm'],
        request_digest,
        self.challenge['qop'],
        self.challenge['nc'],
        self.challenge['cnonce'],
        )
        self.challenge['nc'] += 1

    def response(self, response, content):
        if not response.has_key('authentication-info'):
            challenge = _parse_www_authenticate(response).get('digest', {})
            if 'true' == challenge.get('stale'):
                self.challenge['nonce'] = challenge['nonce']
                self.challenge['nc'] = 1
                return True
        else:
            updated_challenge = _parse_www_authenticate(response, 'authentication-info').get('digest', {})

            if updated_challenge.has_key('nextnonce'):
                self.challenge['nonce'] = updated_challenge['nextnonce']
                self.challenge['nc'] = 1
        return False


class HmacDigestAuthentication(Authentication):
    """Adapted from Robert Sayre's code and DigestAuthentication above."""
    __author__ = "Thomas Broyer (t.broyer@ltgt.net)"

    def __init__(self, credentials, host, request_uri, headers, response, content, http):
        Authentication.__init__(self, credentials, host, request_uri, headers, response, content, http)
        challenge = _parse_www_authenticate(response)
        self.challenge = challenge['hmacdigest']
        # TODO: self.challenge['domain']
        self.challenge['reason'] = self.challenge.get('reason', 'unauthorized')
        if self.challenge['reason'] not in ['unauthorized', 'integrity']:
            self.challenge['reason'] = 'unauthorized'
        self.challenge['salt'] = self.challenge.get('salt', '')
        if not self.challenge.get('snonce'):
            raise UnimplementedHmacDigestAuthOptionError(
                _("The challenge doesn't contain a server nonce, or this one is empty."))
        self.challenge['algorithm'] = self.challenge.get('algorithm', 'HMAC-SHA-1')
        if self.challenge['algorithm'] not in ['HMAC-SHA-1', 'HMAC-MD5']:
            raise UnimplementedHmacDigestAuthOptionError(
                _("Unsupported value for algorithm: %s." % self.challenge['algorithm']))
        self.challenge['pw-algorithm'] = self.challenge.get('pw-algorithm', 'SHA-1')
        if self.challenge['pw-algorithm'] not in ['SHA-1', 'MD5']:
            raise UnimplementedHmacDigestAuthOptionError(
                _("Unsupported value for pw-algorithm: %s." % self.challenge['pw-algorithm']))
        if self.challenge['algorithm'] == 'HMAC-MD5':
            self.hashmod = md5
        else:
            self.hashmod = sha
        if self.challenge['pw-algorithm'] == 'MD5':
            self.pwhashmod = md5
        else:
            self.pwhashmod = sha
        self.key = "".join([self.credentials[0], ":",
                            self.pwhashmod.new(
                                "".join([self.credentials[1], self.challenge['salt']])).hexdigest().lower(),
                            ":", self.challenge['realm']
        ])
        self.key = self.pwhashmod.new(self.key).hexdigest().lower()

    def request(self, method, request_uri, headers, content):
        """Modify the request headers"""
        keys = _get_end2end_headers(headers)
        keylist = "".join(["%s " % k for k in keys])
        headers_val = "".join([headers[k] for k in keys])
        created = time.strftime('%Y-%m-%dT%H:%M:%SZ', time.gmtime())
        cnonce = _cnonce()
        request_digest = "%s:%s:%s:%s:%s" % (method, request_uri, cnonce, self.challenge['snonce'], headers_val)
        request_digest = hmac.new(self.key, request_digest, self.hashmod).hexdigest().lower()
        headers[
        'Authorization'] = 'HMACDigest username="%s", realm="%s", snonce="%s", cnonce="%s", uri="%s", created="%s", response="%s", headers="%s"' % (
        self.credentials[0],
        self.challenge['realm'],
        self.challenge['snonce'],
        cnonce,
        request_uri,
        created,
        request_digest,
        keylist,
        )

    def response(self, response, content):
        challenge = _parse_www_authenticate(response).get('hmacdigest', {})
        if challenge.get('reason') in ['integrity', 'stale']:
            return True
        return False


class WsseAuthentication(Authentication):
    """This is thinly tested and should not be relied upon.
    At this time there isn't any third party server to test against.
    Blogger and TypePad implemented this algorithm at one point
    but Blogger has since switched to Basic over HTTPS and
    TypePad has implemented it wrong, by never issuing a 401
    challenge but instead requiring your client to telepathically know that
    their endpoint is expecting WSSE profile="UsernameToken"."""

    def __init__(self, credentials, host, request_uri, headers, response, content, http):
        Authentication.__init__(self, credentials, host, request_uri, headers, response, content, http)

    def request(self, method, request_uri, headers, content):
        """Modify the request headers to add the appropriate
        Authorization header."""
        headers['Authorization'] = 'WSSE profile="UsernameToken"'
        iso_now = time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime())
        cnonce = _cnonce()
        password_digest = _wsse_username_token(cnonce, iso_now, self.credentials[1])
        headers['X-WSSE'] = 'UsernameToken Username="%s", PasswordDigest="%s", Nonce="%s", Created="%s"' % (
        self.credentials[0],
        password_digest,
        cnonce,
        iso_now)


class GoogleLoginAuthentication(Authentication):
    def __init__(self, credentials, host, request_uri, headers, response, content, http):
        from urllib import urlencode

        Authentication.__init__(self, credentials, host, request_uri, headers, response, content, http)
        challenge = _parse_www_authenticate(response)
        service = challenge['googlelogin'].get('service', 'xapi')
        # Bloggger actually returns the service in the challenge
        # For the rest we guess based on the URI
        if service == 'xapi' and  request_uri.find("calendar") > 0:
            service = "cl"
            # No point in guessing Base or Spreadsheet
        #elif request_uri.find("spreadsheets") > 0:
        #    service = "wise"

        auth = dict(Email=credentials[0], Passwd=credentials[1], service=service, source=headers['user-agent'])
        resp, content = self.http.request("https://www.google.com/accounts/ClientLogin", method="POST",
                                          body=urlencode(auth),
                                          headers={'Content-Type': 'application/x-www-form-urlencoded'})
        lines = content.split('\n')
        d = dict([tuple(line.split("=", 1)) for line in lines if line])
        if resp.status == 403:
            self.Auth = ""
        else:
            self.Auth = d['Auth']

    def request(self, method, request_uri, headers, content):
        """Modify the request headers to add the appropriate
        Authorization header."""
        headers['authorization'] = 'GoogleLogin Auth=' + self.Auth


AUTH_SCHEME_CLASSES = {
    "basic": BasicAuthentication,
    "wsse": WsseAuthentication,
    "digest": DigestAuthentication,
    "hmacdigest": HmacDigestAuthentication,
    "googlelogin": GoogleLoginAuthentication
}

AUTH_SCHEME_ORDER = ["hmacdigest", "googlelogin", "digest", "wsse", "basic"]

def _md5(s):
    return


class FileCache(object):
    """Uses a local directory as a store for cached files.
    Not really safe to use if multiple threads or processes are going to
    be running on the same cache.
    """

    def __init__(self, cache, safe=safename): # use safe=lambda x: md5.new(x).hexdigest() for the old behavior
        self.cache = cache
        self.safe = safe
        if not os.path.exists(cache):
            os.makedirs(self.cache)

    def get(self, key):
        retval = None
        cacheFullPath = os.path.join(self.cache, self.safe(key))
        try:
            f = file(cacheFullPath, "r")
            retval = f.read()
            f.close()
        except IOError:
            pass
        return retval

    def set(self, key, value):
        cacheFullPath = os.path.join(self.cache, self.safe(key))
        f = file(cacheFullPath, "w")
        f.write(value)
        f.close()

    def delete(self, key):
        cacheFullPath = os.path.join(self.cache, self.safe(key))
        if os.path.exists(cacheFullPath):
            os.remove(cacheFullPath)


class Credentials(object):
    def __init__(self):
        self.credentials = []

    def add(self, name, password, domain=""):
        self.credentials.append((domain.lower(), name, password))

    def clear(self):
        self.credentials = []

    def iter(self, domain):
        for (cdomain, name, password) in self.credentials:
            if cdomain == "" or domain == cdomain:
                yield (name, password)


class KeyCerts(Credentials):
    """Identical to Credentials except that
    name/password are mapped to key/cert."""
    pass


class ProxyInfo(object):
    """Collect information required to use a proxy."""

    def __init__(self, proxy_type, proxy_host, proxy_port, proxy_rdns=None, proxy_user=None, proxy_pass=None):
        """The parameter proxy_type must be set to one of socks.PROXY_TYPE_XXX
        constants. For example:

  p = ProxyInfo(proxy_type=socks.PROXY_TYPE_HTTP, proxy_host='localhost', proxy_port=8000)
        """
        self.proxy_type, self.proxy_host, self.proxy_port, self.proxy_rdns, self.proxy_user, self.proxy_pass = proxy_type, proxy_host, proxy_port, proxy_rdns, proxy_user, proxy_pass

    def astuple(self):
        return (self.proxy_type, self.proxy_host, self.proxy_port, self.proxy_rdns,
                self.proxy_user, self.proxy_pass)

    def isgood(self):
        return socks and (self.proxy_host != None) and (self.proxy_port != None)


class HTTPConnectionWithTimeout(httplib.HTTPConnection):
    """HTTPConnection subclass that supports timeouts"""

    def __init__(self, host, port=None, strict=None, timeout=None, proxy_info=None):
        httplib.HTTPConnection.__init__(self, host, port, strict)
        self.timeout = timeout
        self.proxy_info = proxy_info

    def connect(self):
        """Connect to the host and port specified in __init__."""
        # Mostly verbatim from httplib.py.
        msg = "getaddrinfo returns an empty list"
        for res in socket.getaddrinfo(self.host, self.port, 0,
                                      socket.SOCK_STREAM):
            af, socktype, proto, canonname, sa = res
            try:
                if self.proxy_info and self.proxy_info.isgood():
                    self.sock = socks.socksocket(af, socktype, proto)
                    self.sock.setproxy(*self.proxy_info.astuple())
                else:
                    self.sock = socket.socket(af, socktype, proto)
                    # Different from httplib: support timeouts.
                if self.timeout is not None:
                    self.sock.settimeout(self.timeout)
                    # End of difference from httplib.
                if self.debuglevel > 0:
                    print "connect: (%s, %s)" % (self.host, self.port)
                self.sock.connect(sa)
            except socket.error, msg:
                if self.debuglevel > 0:
                    print 'connect fail:', (self.host, self.port)
                if self.sock:
                    self.sock.close()
                self.sock = None
                continue
            break
        if not self.sock:
            raise socket.error, msg


class HTTPSConnectionWithTimeout(httplib.HTTPSConnection):
    "This class allows communication via SSL."

    def __init__(self, host, port=None, key_file=None, cert_file=None,
                 strict=None, timeout=None, proxy_info=None):
        self.timeout = timeout
        self.proxy_info = proxy_info
        httplib.HTTPSConnection.__init__(self, host, port=port, key_file=key_file,
                                         cert_file=cert_file, strict=strict)

    def connect(self):
        "Connect to a host on a given (SSL) port."

        if self.proxy_info and self.proxy_info.isgood():
            self.sock.setproxy(*self.proxy_info.astuple())
            sock.setproxy(*self.proxy_info.astuple())
        else:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        if self.timeout is not None:
            sock.settimeout(self.timeout)
        sock.connect((self.host, self.port))
        ssl = socket.ssl(sock, self.key_file, self.cert_file)
        self.sock = httplib.FakeSocket(sock, ssl)


class Http(object):
    """An HTTP client that handles:
- all methods
- caching
- ETags
- compression,
- HTTPS
- Basic
- Digest
- WSSE

and more.
    """

    def __init__(self, cache=None, timeout=None, proxy_info=None):
        """The value of proxy_info is a ProxyInfo instance.

If 'cache' is a string then it is used as a directory name
for a disk cache. Otherwise it must be an object that supports
the same interface as FileCache."""
        self.proxy_info = proxy_info
        # Map domain name to an httplib connection
        self.connections = {}
        # The location of the cache, for now a directory
        # where cached responses are held.
        if cache and isinstance(cache, str):
            self.cache = FileCache(cache)
        else:
            self.cache = cache

        # Name/password
        self.credentials = Credentials()

        # Key/cert
        self.certificates = KeyCerts()

        # authorization objects
        self.authorizations = []

        # If set to False then no redirects are followed, even safe ones.
        self.follow_redirects = True

        # If 'follow_redirects' is True, and this is set to True then
        # all redirecs are followed, including unsafe ones.
        self.follow_all_redirects = False

        self.ignore_etag = False

        self.force_exception_to_status_code = False

        self.timeout = timeout

    def _auth_from_challenge(self, host, request_uri, headers, response, content):
        """A generator that creates Authorization objects
           that can be applied to requests.
        """
        challenges = _parse_www_authenticate(response)
        for cred in self.credentials.iter(host):
            for scheme in AUTH_SCHEME_ORDER:
                if challenges.has_key(scheme):
                    yield AUTH_SCHEME_CLASSES[scheme](cred, host, request_uri, headers, response, content, self)

    def add_credentials(self, name, password, domain=""):
        """Add a name and password that will be used
        any time a request requires authentication."""
        self.credentials.add(name, password, domain)

    def add_certificate(self, key, cert, domain):
        """Add a key and cert that will be used
        any time a request requires authentication."""
        self.certificates.add(key, cert, domain)

    def clear_credentials(self):
        """Remove all the names and passwords
        that are used for authentication"""
        self.credentials.clear()
        self.authorizations = []

    def _conn_request(self, conn, request_uri, method, body, headers):
        for i in range(2):
            try:
                conn.request(method, request_uri, body, headers)
                response = conn.getresponse()
            except socket.gaierror:
                conn.close()
                raise ServerNotFoundError("Unable to find the server at %s" % conn.host)
            except httplib.HTTPException, e:
                if not i:
                    conn.close()
                    conn.connect()
                    continue
                else:
                    raise
            else:
                content = response.read()
                response = Response(response)
                if method != "HEAD":
                    content = _decompressContent(response, content)

            break
        return (response, content)


    def _request(self, conn, host, absolute_uri, request_uri, method, body, headers, redirections, cachekey):
        """Do the actual request using the connection object
        and also follow one level of redirects if necessary"""

        auths = [(auth.depth(request_uri), auth) for auth in self.authorizations if auth.inscope(host, request_uri)]
        auth = auths and sorted(auths)[0][1] or None
        if auth:
            auth.request(method, request_uri, headers, body)

        (response, content) = self._conn_request(conn, request_uri, method, body, headers)

        if auth:
            if auth.response(response, body):
                auth.request(method, request_uri, headers, body)
                (response, content) = self._conn_request(conn, request_uri, method, body, headers)
                response._stale_digest = 1

        if response.status == 401:
            for authorization in self._auth_from_challenge(host, request_uri, headers, response, content):
                authorization.request(method, request_uri, headers, body)
                (response, content) = self._conn_request(conn, request_uri, method, body, headers, )
                if response.status != 401:
                    self.authorizations.append(authorization)
                    authorization.response(response, body)
                    break

        if (self.follow_all_redirects or (method in ["GET", "HEAD"]) or response.status == 303):
            if self.follow_redirects and response.status in [300, 301, 302, 303, 307]:
                # Pick out the location header and basically start from the beginning
                # remembering first to strip the ETag header and decrement our 'depth'
                if redirections:
                    if not response.has_key('location') and response.status != 300:
                        raise RedirectMissingLocation(_("Redirected but the response is missing a Location: header."),
                                                      response, content)
                        # Fix-up relative redirects (which violate an RFC 2616 MUST)
                    if response.has_key('location'):
                        location = response['location']
                        (scheme, authority, path, query, fragment) = parse_uri(location)
                        if authority == None:
                            response['location'] = urlparse.urljoin(absolute_uri, location)
                    if response.status == 301 and method in ["GET", "HEAD"]:
                        response['-x-permanent-redirect-url'] = response['location']
                        if not response.has_key('content-location'):
                            response['content-location'] = absolute_uri
                        _updateCache(headers, response, content, self.cache, cachekey)
                    if headers.has_key('if-none-match'):
                        del headers['if-none-match']
                    if headers.has_key('if-modified-since'):
                        del headers['if-modified-since']
                    if response.has_key('location'):
                        location = response['location']
                        old_response = copy.deepcopy(response)
                        if not old_response.has_key('content-location'):
                            old_response['content-location'] = absolute_uri
                        redirect_method = ((response.status == 303) and (
                        method not in ["GET", "HEAD"])) and "GET" or method
                        (response, content) = self.request(location, redirect_method, body=body, headers=headers,
                                                           redirections=redirections - 1)
                        response.previous = old_response
                else:
                    raise RedirectLimit(_("Redirected more times than rediection_limit allows."), response, content)
            elif response.status in [200, 203] and method == "GET":
                # Don't cache 206's since we aren't going to handle byte range requests
                if not response.has_key('content-location'):
                    response['content-location'] = absolute_uri
                _updateCache(headers, response, content, self.cache, cachekey)

        return (response, content)


    # Need to catch and rebrand some exceptions
    # Then need to optionally turn all exceptions into status codes
    # including all socket.* and httplib.* exceptions.


    def request(self, uri, method="GET", body=None, headers=None, redirections=DEFAULT_MAX_REDIRECTS,
                connection_type=None):
        """ Performs a single HTTP request.
The 'uri' is the URI of the HTTP resource and can begin
with either 'http' or 'https'. The value of 'uri' must be an absolute URI.

The 'method' is the HTTP method to perform, such as GET, POST, DELETE, etc.
There is no restriction on the methods allowed.

The 'body' is the entity body to be sent with the request. It is a string
object.

Any extra headers that are to be sent with the request should be provided in the
'headers' dictionary.

The maximum number of redirect to follow before raising an
exception is 'redirections. The default is 5.

The return value is a tuple of (response, content), the first
being and instance of the 'Response' class, the second being
a string that contains the response entity body.
        """
        try:
            if headers is None:
                headers = {}
            else:
                headers = _normalize_headers(headers)

            if not headers.has_key('user-agent'):
                headers['user-agent'] = "Python-httplib2/%s" % __version__

            uri = iri2uri(uri)

            (scheme, authority, request_uri, defrag_uri) = urlnorm(uri)

            conn_key = scheme + ":" + authority
            if conn_key in self.connections:
                conn = self.connections[conn_key]
            else:
                if not connection_type:
                    connection_type = (scheme == 'https') and HTTPSConnectionWithTimeout or HTTPConnectionWithTimeout
                certs = list(self.certificates.iter(authority))
                if scheme == 'https' and certs:
                    conn = self.connections[conn_key] = connection_type(authority, key_file=certs[0][0],
                                                                        cert_file=certs[0][1], timeout=self.timeout,
                                                                        proxy_info=self.proxy_info)
                else:
                    conn = self.connections[conn_key] = connection_type(authority, timeout=self.timeout,
                                                                        proxy_info=self.proxy_info)
                conn.set_debuglevel(debuglevel)

            if method in ["GET", "HEAD"] and 'range' not in headers:
                headers['accept-encoding'] = 'compress, gzip'

            info = email.Message.Message()
            cached_value = None
            if self.cache:
                cachekey = defrag_uri
                cached_value = self.cache.get(cachekey)
                if cached_value:
                    info = email.message_from_string(cached_value)
                    try:
                        content = cached_value.split('\r\n\r\n', 1)[1]
                    except IndexError:
                        self.cache.delete(cachekey)
                        cachekey = None
                        cached_value = None
            else:
                cachekey = None

            if method in ["PUT"] and self.cache and info.has_key(
                'etag') and not self.ignore_etag and 'if-match' not in headers:
                # http://www.w3.org/1999/04/Editing/
                headers['if-match'] = info['etag']

            if method not in ["GET", "HEAD"] and self.cache and cachekey:
                # RFC 2616 Section 13.10
                self.cache.delete(cachekey)

            if cached_value and method in ["GET", "HEAD"] and self.cache and 'range' not in headers:
                if info.has_key('-x-permanent-redirect-url'):
                    # Should cached permanent redirects be counted in our redirection count? For now, yes.
                    (response, new_content) = self.request(info['-x-permanent-redirect-url'], headers=headers,
                                                           redirections=redirections - 1)
                    response.previous = Response(info)
                    response.previous.fromcache = True
                else:
                    # Determine our course of action:
                    #   Is the cached entry fresh or stale?
                    #   Has the client requested a non-cached response?
                    #
                    # There seems to be three possible answers:
                    # 1. [FRESH] Return the cache entry w/o doing a GET
                    # 2. [STALE] Do the GET (but add in cache validators if available)
                    # 3. [TRANSPARENT] Do a GET w/o any cache validators (Cache-Control: no-cache) on the request
                    entry_disposition = _entry_disposition(info, headers)

                    if entry_disposition == "FRESH":
                        if not cached_value:
                            info['status'] = '504'
                            content = ""
                        response = Response(info)
                        if cached_value:
                            response.fromcache = True
                        return (response, content)

                    if entry_disposition == "STALE":
                        if info.has_key('etag') and not self.ignore_etag and not 'if-none-match' in headers:
                            headers['if-none-match'] = info['etag']
                        if info.has_key('last-modified') and not 'last-modified' in headers:
                            headers['if-modified-since'] = info['last-modified']
                    elif entry_disposition == "TRANSPARENT":
                        pass

                    (response, new_content) = self._request(conn, authority, uri, request_uri, method, body, headers,
                                                            redirections, cachekey)

                if response.status == 304 and method == "GET":
                    # Rewrite the cache entry with the new end-to-end headers
                    # Take all headers that are in response 
                    # and overwrite their values in info.
                    # unless they are hop-by-hop, or are listed in the connection header.

                    for key in _get_end2end_headers(response):
                        info[key] = response[key]
                    merged_response = Response(info)
                    if hasattr(response, "_stale_digest"):
                        merged_response._stale_digest = response._stale_digest
                    _updateCache(headers, merged_response, content, self.cache, cachekey)
                    response = merged_response
                    response.status = 200
                    response.fromcache = True

                elif response.status == 200:
                    content = new_content
                else:
                    self.cache.delete(cachekey)
                    content = new_content
            else:
                (response, content) = self._request(conn, authority, uri, request_uri, method, body, headers,
                                                    redirections, cachekey)
        except Exception, e:
            if self.force_exception_to_status_code:
                if isinstance(e, HttpLib2ErrorWithResponse):
                    response = e.response
                    content = e.content
                    response.status = 500
                    response.reason = str(e)
                elif isinstance(e, socket.timeout):
                    content = "Request Timeout"
                    response = Response({
                        "content-type": "text/plain",
                        "status": "408",
                        "content-length": len(content)
                    })
                    response.reason = "Request Timeout"
                else:
                    content = str(e)
                    response = Response({
                        "content-type": "text/plain",
                        "status": "400",
                        "content-length": len(content)
                    })
                    response.reason = "Bad Request"
            else:
                raise

        return (response, content)


class Response(dict):
    """An object more like email.Message than httplib.HTTPResponse."""

    """Is this response from our local cache"""
    fromcache = False

    """HTTP protocol version used by server. 10 for HTTP/1.0, 11 for HTTP/1.1. """
    version = 11

    "Status code returned by server. "
    status = 200

    """Reason phrase returned by server."""
    reason = "Ok"

    previous = None

    def __init__(self, info):
        # info is either an email.Message or
        # an httplib.HTTPResponse object.
        if isinstance(info, httplib.HTTPResponse):
            for key, value in info.getheaders():
                self[key] = value
            self.status = info.status
            self['status'] = str(self.status)
            self.reason = info.reason
            self.version = info.version
        elif isinstance(info, email.Message.Message):
            for key, value in info.items():
                self[key] = value
            self.status = int(self['status'])
        else:
            for key, value in info.iteritems():
                self[key] = value
            self.status = int(self.get('status', self.status))


    def __getattr__(self, name):
        if name == 'dict':
            return self
        else:
            raise AttributeError, name



#!/usr/bin/env python

import re
import time

from section import Section

class Table(object):
    """
    A table contains several @class Section

    - Horizontal view

        section-1
        ---------
        section-2
        ---------
        section-3
    """
    def __init__(self, name="", ftr="", width=800, height=600, sep="-"):
        self.name = name
        self.ftr = ftr
        self.width = width          # max width
        self.height = height        # max height
        self.sep = sep
        self.sections = {}          # @dict {name: @class Section}
        self.hori = True

    def __str__(self):
        div = "%s\n" % (self.sep * self.size()[0])
        return div.join(map(lambda x: str(x),
                            sorted(self.sections.itervalues(),
                                   key=lambda y: y.id)) + [self.get_ftr()])

    def __repr__(self):
        return "Table(name=%r, sections=%r, hori=%r)"\
            % (self.name, self.sections, self.hori)

    def size(self):
        """Return the viewable size of the Table as @tuple (x,y)"""
        width = max(
            map(lambda x: x.size()[0], self.sections.itervalues()))

        height = sum(
            map(lambda x: x.size()[1], self.sections.itervalues()))

        return width, height

    def set_ftr(self, ftr):
        """
        Set footer for this table
        Accept the following regexps:

        %time : current time printed as "hh:mm:ss"
        """
        self.ftr = ftr

    def get_ftr(self):
        """
        Process footer and return the processed string
        """
        if not self.ftr:
            return self.ftr

        width = self.size()[0]
        return re.sub(
            "%time", "%s\n" % time.strftime("%H:%M:%S"), self.ftr).rjust(width)

    def get_section(self, name):
        if name in self.sections.iterkeys():
            return self.sections[name]
        return None

    def add_section(self, section):
        if isinstance(section, Section):
            self.sections[section.name] = section


# Copyright 2009-2010 10gen, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tools for creating `messages
<http://www.mongodb.org/display/DOCS/Mongo+Wire+Protocol>`_ to be sent to
MongoDB.

.. note:: This module is for internal use and is generally not needed by
   application developers.

.. versionadded:: 1.1.2
"""

import random
import struct

import bson
from bson.binary import OLD_UUID_SUBTYPE
from bson.son import SON
try:
    from pymongo import _cmessage
    _use_c = True
except ImportError:
    _use_c = False
from pymongo.errors import InvalidOperation


__ZERO = "\x00\x00\x00\x00"

MAX_INT32 = 2147483647
MIN_INT32 = -2147483648


def __last_error(args):
    """Data to send to do a lastError.
    """
    cmd = SON([("getlasterror", 1)])
    cmd.update(args)
    return query(0, "admin.$cmd", 0, -1, cmd)


def __pack_message(operation, data):
    """Takes message data and adds a message header based on the operation.

    Returns the resultant message string.
    """
    request_id = random.randint(MIN_INT32, MAX_INT32)
    message = struct.pack("<i", 16 + len(data))
    message += struct.pack("<i", request_id)
    message += __ZERO  # responseTo
    message += struct.pack("<i", operation)
    return (request_id, message + data)


def insert(collection_name, docs, check_keys,
           safe, last_error_args, continue_on_error, uuid_subtype):
    """Get an **insert** message.
    """
    max_bson_size = 0
    options = 0
    if continue_on_error:
        options += 1
    data = struct.pack("<i", options)
    data += bson._make_c_string(collection_name)
    encoded = [bson.BSON.encode(doc, check_keys, uuid_subtype) for doc in docs]
    if not encoded:
        raise InvalidOperation("cannot do an empty bulk insert")
    max_bson_size = max(map(len, encoded))
    data += "".join(encoded)
    if safe:
        (_, insert_message) = __pack_message(2002, data)
        (request_id, error_message, _) = __last_error(last_error_args)
        return (request_id, insert_message + error_message, max_bson_size)
    else:
        (request_id, insert_message) = __pack_message(2002, data)
        return (request_id, insert_message, max_bson_size)
if _use_c:
    insert = _cmessage._insert_message


def update(collection_name, upsert, multi,
           spec, doc, safe, last_error_args, check_keys, uuid_subtype):
    """Get an **update** message.
    """
    options = 0
    if upsert:
        options += 1
    if multi:
        options += 2

    data = __ZERO
    data += bson._make_c_string(collection_name)
    data += struct.pack("<i", options)
    data += bson.BSON.encode(spec, False, uuid_subtype)
    encoded = bson.BSON.encode(doc, check_keys, uuid_subtype)
    data += encoded
    if safe:
        (_, update_message) = __pack_message(2001, data)
        (request_id, error_message, _) = __last_error(last_error_args)
        return (request_id, update_message + error_message, len(encoded))
    else:
        (request_id, update_message) = __pack_message(2001, data)
        return (request_id, update_message, len(encoded))
if _use_c:
    update = _cmessage._update_message


def query(options, collection_name, num_to_skip,
          num_to_return, query, field_selector=None,
          uuid_subtype=OLD_UUID_SUBTYPE):
    """Get a **query** message.
    """
    data = struct.pack("<I", options)
    data += bson._make_c_string(collection_name)
    data += struct.pack("<i", num_to_skip)
    data += struct.pack("<i", num_to_return)
    encoded = bson.BSON.encode(query, False, uuid_subtype)
    data += encoded
    max_bson_size = len(encoded)
    if field_selector is not None:
        encoded = bson.BSON.encode(field_selector, False, uuid_subtype)
        data += encoded
        max_bson_size = max(len(encoded), max_bson_size)
    (request_id, query_message) = __pack_message(2004, data)
    return (request_id, query_message, max_bson_size)
if _use_c:
    query = _cmessage._query_message


def get_more(collection_name, num_to_return, cursor_id):
    """Get a **getMore** message.
    """
    data = __ZERO
    data += bson._make_c_string(collection_name)
    data += struct.pack("<i", num_to_return)
    data += struct.pack("<q", cursor_id)
    return __pack_message(2005, data)
if _use_c:
    get_more = _cmessage._get_more_message


def delete(collection_name, spec, safe, last_error_args, uuid_subtype):
    """Get a **delete** message.
    """
    data = __ZERO
    data += bson._make_c_string(collection_name)
    data += __ZERO
    encoded = bson.BSON.encode(spec, False, uuid_subtype)
    data += encoded
    if safe:
        (_, remove_message) = __pack_message(2006, data)
        (request_id, error_message, _) = __last_error(last_error_args)
        return (request_id, remove_message + error_message, len(encoded))
    else:
        (request_id, remove_message) = __pack_message(2006, data)
        return (request_id, remove_message, len(encoded))


def kill_cursors(cursor_ids):
    """Get a **killCursors** message.
    """
    data = __ZERO
    data += struct.pack("<i", len(cursor_ids))
    for cursor_id in cursor_ids:
        data += struct.pack("<q", cursor_id)
    return __pack_message(2007, data)



class WhisperException(Exception):
    """Base class for whisper exceptions."""


class InvalidConfiguration(WhisperException):
    """Invalid configuration."""


class InvalidAggregationMethod(WhisperException):
    """Invalid aggregation method."""


class InvalidTimeInterval(WhisperException):
    """Invalid time interval."""


class TimestampNotCovered(WhisperException):
    """Timestamp not covered by any archives in this database."""


class CorruptWhisperFile(WhisperException):
    def __init__(self, error, path):
        super(CorruptWhisperFile, self).__init__(self, error)
        self.error = error
        self.path = path

    def __repr__(self):
        return "<CorruptWhisperFile[%s] %s>" % (self.path, self.error)

    def __str__(self):
        return "%s (%s)" % (self.error, self.path)



# Copyright 2009-2010 10gen, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Manipulators that can edit SON objects as they enter and exit a database.

New manipulators should be defined as subclasses of SONManipulator and can be
installed on a database by calling
`pymongo.database.Database.add_son_manipulator`."""

from bson.dbref import DBRef
from bson.objectid import ObjectId
from bson.son import SON


class SONManipulator(object):
    """A base son manipulator.

    This manipulator just saves and restores objects without changing them.
    """

    def will_copy(self):
        """Will this SON manipulator make a copy of the incoming document?

        Derived classes that do need to make a copy should override this
        method, returning True instead of False. All non-copying manipulators
        will be applied first (so that the user's document will be updated
        appropriately), followed by copying manipulators.
        """
        return False

    def transform_incoming(self, son, collection):
        """Manipulate an incoming SON object.

        :Parameters:
          - `son`: the SON object to be inserted into the database
          - `collection`: the collection the object is being inserted into
        """
        if self.will_copy():
            return SON(son)
        return son

    def transform_outgoing(self, son, collection):
        """Manipulate an outgoing SON object.

        :Parameters:
          - `son`: the SON object being retrieved from the database
          - `collection`: the collection this object was stored in
        """
        if self.will_copy():
            return SON(son)
        return son


class ObjectIdInjector(SONManipulator):
    """A son manipulator that adds the _id field if it is missing.
    """

    def transform_incoming(self, son, collection):
        """Add an _id field if it is missing.
        """
        if not "_id" in son:
            son["_id"] = ObjectId()
        return son


# This is now handled during BSON encoding (for performance reasons),
# but I'm keeping this here as a reference for those implementing new
# SONManipulators.
class ObjectIdShuffler(SONManipulator):
    """A son manipulator that moves _id to the first position.
    """

    def will_copy(self):
        """We need to copy to be sure that we are dealing with SON, not a dict.
        """
        return True

    def transform_incoming(self, son, collection):
        """Move _id to the front if it's there.
        """
        if not "_id" in son:
            return son
        transformed = SON({"_id": son["_id"]})
        transformed.update(son)
        return transformed


class NamespaceInjector(SONManipulator):
    """A son manipulator that adds the _ns field.
    """

    def transform_incoming(self, son, collection):
        """Add the _ns field to the incoming object
        """
        son["_ns"] = collection.name
        return son


class AutoReference(SONManipulator):
    """Transparently reference and de-reference already saved embedded objects.

    This manipulator should probably only be used when the NamespaceInjector is
    also being used, otherwise it doesn't make too much sense - documents can
    only be auto-referenced if they have an *_ns* field.

    NOTE: this will behave poorly if you have a circular reference.

    TODO: this only works for documents that are in the same database. To fix
    this we'll need to add a DatabaseInjector that adds *_db* and then make
    use of the optional *database* support for DBRefs.
    """

    def __init__(self, db):
        self.__database = db

    def will_copy(self):
        """We need to copy so the user's document doesn't get transformed refs.
        """
        return True

    def transform_incoming(self, son, collection):
        """Replace embedded documents with DBRefs.
        """

        def transform_value(value):
            if isinstance(value, dict):
                if "_id" in value and "_ns" in value:
                    return DBRef(value["_ns"], transform_value(value["_id"]))
                else:
                    return transform_dict(SON(value))
            elif isinstance(value, list):
                return [transform_value(v) for v in value]
            return value

        def transform_dict(object):
            for (key, value) in object.items():
                object[key] = transform_value(value)
            return object

        return transform_dict(SON(son))

    def transform_outgoing(self, son, collection):
        """Replace DBRefs with embedded documents.
        """

        def transform_value(value):
            if isinstance(value, DBRef):
                return self.__database.dereference(value)
            elif isinstance(value, list):
                return [transform_value(v) for v in value]
            elif isinstance(value, dict):
                return transform_dict(SON(value))
            return value

        def transform_dict(object):
            for (key, value) in object.items():
                object[key] = transform_value(value)
            return object

        return transform_dict(SON(son))

# TODO make a generic translator for custom types. Take encode, decode,
# should_encode and should_decode functions and just encode and decode where
# necessary. See examples/custom_type.py for where this would be useful.
# Alternatively it could take a should_encode, to_binary, from_binary and
# binary subtype.



import json
import os
import re
from threading import Thread
import time
from membase.api.rest_client import RestConnection
from memcached.helper.data_helper import MemcachedClientHelper
from remote.remote_util import RemoteMachineShellConnection, RemoteMachineHelper
import testconstants
import gzip
import urllib

from mc_bin_client import MemcachedError

RETRIES = 10
# The histo dict is returned by add_timing_sample().
# The percentiles must be sorted, ascending, like [0.90, 0.99].
def histo_percentile(histo, percentiles):
   v_sum = 0
   bins = histo.keys()
   bins.sort()
   for bin in bins:
      v_sum += histo[bin]
   v_sum = float(v_sum)
   v_cur = 0 # Running total.
   rv = []
   for bin in bins:
      if not percentiles:
         return rv
      v_cur += histo[bin]
      while percentiles and (v_cur / v_sum) >= percentiles[0]:
         rv.append((percentiles[0], bin))
         percentiles.pop(0)
   return rv

class StatsCollector(object):
    _task = {}
    _verbosity = True
    _mb_stats = {"snapshots": []}   # manually captured memcached stats
    _reb_stats = {}

    def __init__(self, verbosity):
        self._verbosity = verbosity
        self.is_leader = False
        self.active_mergers = 0

    #this function starts collecting stats from all nodes with the given
    #frequency
    def start(self, nodes, bucket, pnames, name, frequency, client_id='',
              collect_server_stats = True):
        self._task = {"state": "running", "threads": []}
        self._task["name"] = name
        self._task["time"] = time.time()
        self._task["ops"] = []
        self._task["totalops"] = []
        self._task["ops-temp"] = []
        self._task["latency"] = {}
        self._task["data_size_stats"] = []
        rest = RestConnection(nodes[0])
        info = rest.get_nodes_self()
        self.data_path = info.storage[0].get_data_path()
        self.client_id = str(client_id)

        if collect_server_stats:
            mbstats_thread = Thread(target=self.membase_stats,
                                    args=(nodes, bucket, 600, self._verbosity))
            mbstats_thread.start()
            sysstats_thread = Thread(target=self.system_stats,
                                     args=(nodes, pnames, frequency, self._verbosity))
            sysstats_thread.start()
            ns_server_stats_thread = Thread(target=self.ns_server_stats,
                                            args=([nodes[0]], bucket, 60, self._verbosity))
            ns_server_stats_thread.start()
            rest = RestConnection(nodes[0])
            bucket_size_thead = Thread(target=self.get_bucket_size,
                                       args=(bucket, rest, frequency))
            bucket_size_thead.start()
#            data_size_thread = Thread(target=self.get_data_file_size,
#                                    args=(nodes, 60, bucket))
#            data_size_thread.start()
            self._task["threads"] = [sysstats_thread, ns_server_stats_thread, bucket_size_thead,
                                     mbstats_thread]
                                     #data_size_thread ]
            # Getting build/machine stats from only one node in the cluster
            self.build_stats(nodes)
            self.machine_stats(nodes)

    def stop(self):
        self._task["state"] = "stopped"
        for t in self._task["threads"]:
            t.join()

        self._task["time"] = time.time() - self._task["time"]

    def sample(self, cur):
        pass

    def export(self, name, test_params):
        for latency in self._task["latency"].keys():

            # save the last histogram snapshot
            histos = self._task["latency"].get(latency, [])

            if histos:
                key = latency + "-histogram"
                self._task["latency"][key] = histos[-1].copy()
                del self._task["latency"][key]["delta"]
                self._task["latency"][key]["client_id"] = self.client_id

            # calculate percentiles
            key = 'percentile-' + latency
            self._task["latency"][key] = []
            for histo in histos:
                # for every sample histogram, produce a temp summary:
                # temp = [90 per, 95 per, 99 per, client_id, delta]
                temp = []
                time = histo['time']
                delta = histo['delta']
                del histo['delta'], histo['time']
                p = histo_percentile(histo, [0.80, 0.90, 0.95, 0.99, 0.999])
                # p is list of tuples
                for val in p:
                    temp.append(val[-1])
                temp.append(self.client_id)
                temp.append(time)
                temp.append(delta)
                self._task["latency"][key].append(temp)

        test_params.update(self._reb_stats)

        obj = {"buildinfo": self._task.get("buildstats", {}),
               "machineinfo": self._task.get("machinestats", {}),
               "membasestats": self._task.get("membasestats", []),
               "systemstats": self._task.get("systemstats", []),
               "name": name,
               "totalops":self._task["totalops"],
               "ops":self._task["ops"],
               "time": self._task["time"],
               "info": test_params,
               "ns_server_data": self._task.get("ns_server_stats", []),
               "ns_server_data_system": self._task.get("ns_server_stats_system", []),
               "timings": self._task.get("timings", []),
               "dispatcher": self._task.get("dispatcher", []),
               "bucket-size":self._task.get("bucket_size", []),
               "data-size": self._task.get("data_size_stats", []),
               "latency-set-histogram":self._task["latency"].get("latency-set-histogram", []),
               "latency-set":self._task["latency"].get('percentile-latency-set', []),
               "latency-set-recent":self._task["latency"].get('percentile-latency-set-recent', []),
               "latency-get-histogram":self._task["latency"].get("latency-get-histogram", []),
               "latency-get":self._task["latency"].get('percentile-latency-get', []),
               "latency-get-recent":self._task["latency"].get('percentile-latency-get-recent', []),
               "latency-delete":self._task["latency"].get('percentile-latency-delete', []),
               "latency-delete-recent":self._task["latency"].get('percentile-latency-delete-recent', []),
               "latency-query-histogram":self._task["latency"].get("latency-query-histogram", []),
               "latency-query":self._task["latency"].get('percentile-latency-query', []),
               "latency-query-recent":self._task["latency"].get('percentile-latency-query-recent', []),
               "latency-obs-persist-server-histogram":self._task["latency"].get("latency-obs-persist-server-histogram", []),
               "latency-obs-persist-server":self._task["latency"].get('percentile-latency-obs-persist-server-server', []),
               "latency-obs-persist-server-recent":self._task["latency"].get('percentile-latency-obs-persist-server-recent', []),
               "latency-obs-persist-client-histogram":self._task["latency"].get("latency-obs-persist-client-histogram", []),
               "latency-obs-persist-client":self._task["latency"].get('percentile-latency-obs-persist-client', []),
               "latency-obs-persist-client-recent":self._task["latency"].get('percentile-latency-obs-persist-client-recent', []),
               "latency-obs-repl-client-histogram":self._task["latency"].get("latency-obs-repl-client-histogram", []),
               "latency-obs-repl-client":self._task["latency"].get('percentile-latency-obs-repl-client', []),
               "latency-obs-repl-client-recent":self._task["latency"].get('percentile-latency-obs-repl-client-recent', []),
               "latency-woq-obs-histogram":self._task["latency"].get("latency-woq-obs-histogram", []),
               "latency-woq-obs":self._task["latency"].get('percentile-latency-woq-obs', []),
               "latency-woq-obs-recent":self._task["latency"].get('percentile-latency-woq-obs-recent', []),
               "latency-woq-query-histogram":self._task["latency"].get("latency-woq-query-histogram", []),
               "latency-woq-query":self._task["latency"].get('percentile-latency-woq-query', []),
               "latency-woq-query-recent":self._task["latency"].get('percentile-latency-woq-query-recent', []),
               "latency-woq-histogram":self._task["latency"].get("latency-woq-histogram", []),
               "latency-woq":self._task["latency"].get('percentile-latency-woq', []),
               "latency-woq-recent":self._task["latency"].get('percentile-latency-woq-recent', [])
               }

        if self.client_id:
            patterns = ['reload$', 'load$', 'warmup$', 'index$']
            phases = ['.reload', '.load', '.warmup', '.index']

            name_picker = lambda (pattern, phase): re.search(pattern, self._task["name"])
            try:
                phase = filter(name_picker, zip(patterns, phases))[0][1]
            except IndexError:
                phase = '.loop'

            filename = str(self.client_id) + phase

            file = gzip.open("{0}.json.gz".format(filename), 'wb')
            file.write("{0}".format(json.dumps(obj)))
            file.close()
        else:
            file = gzip.open("{0}.json.gz".format(name), 'wb')
            file.write("{0}".format(json.dumps(obj)))
            file.close()

    def get_bucket_size(self, bucket, rest, frequency):
        self._task["bucket_size"] = []
        d = []
        while not self._aborted():
            print "Collecting bucket size stats"
            status, db_size = rest.get_database_disk_size(bucket)
            if status:
                d.append(db_size)
            else:
                print "Enable to read bucket stats"
            time.sleep(frequency)

        self._task["bucket_size"] = d
        print "finished bucket size stats"

    def get_data_file_size(self, nodes, frequency, bucket):
        shells = []
        for node in nodes:
            try:
                shells.append(RemoteMachineShellConnection(node))
            except:
                pass
        paths = []
        if shells[0].is_membase_installed():
            paths.append(self.data_path+'/{0}-data'.format(bucket))
        else:
            bucket_path = self.data_path+'/{0}'.format(bucket)
            paths.append(bucket_path)
            view_path = bucket_path +'/set_view_{0}_design'.format(bucket)
            paths.append(view_path)

        d = {"snapshots": []}
        start_time = str(self._task["time"])

        while not self._aborted():
            time.sleep(frequency)
            current_time = time.time()
            i = 0
            for shell in shells:
                node = nodes[i]
                unique_id = node.ip+'-'+start_time
                value = {}
                for path in paths:
                    size = shell.get_data_file_size(path)
                    value["file"] = path.split('/')[-1]
                    value["size"] = size
                    value["unique_id"] = unique_id
                    value["time"] = current_time
                    value["ip"] = node.ip
                    d["snapshots"].append(value.copy())
                i +=  1
        self._task["data_size_stats"] = d["snapshots"]
        print " finished data_size_stats"

    #ops stats
    #{'tot-sets': 899999, 'tot-gets': 1, 'tot-items': 899999, 'tot-creates': 899999}
    def ops_stats(self, ops_stat):
        ops_stat["time"] = time.time()
        self._task["ops-temp"].append(ops_stat)
        if len(self._task["ops-temp"]) >= 500 * (1 + self.active_mergers):
            # Prevent concurrent merge
            while self.active_mergers:
                time.sleep(0.1)

            # Semaphore: +1 active
            self.active_mergers += 1

            # Merge
            merged = self._merge()
            self._task["ops"].append(merged)
            self._task["ops-temp"] = self._task["ops-temp"][500:]

            # Semaphore: -1 active
            self.active_mergers -= 1

        #if self._task["ops"] has more than 1000 elements try to aggregate them ?

    def latency_stats(self, latency_cmd, latency_stat, cur_time=0):
        if self._task["latency"].get(latency_cmd) is None:
            self._task["latency"][latency_cmd] = []
        temp_latency_stat = latency_stat.copy()
        if not cur_time:
            cur_time = time.time()
        temp_latency_stat['time'] = int(cur_time)
        temp_latency_stat['delta'] = cur_time - self._task['time']
        self._task["latency"][latency_cmd].append(temp_latency_stat)

    def _merge(self):
        first = self._task["ops-temp"][0]
        merged = {"startTime":first["start-time"]}
        totalgets = 0
        totalsets = 0
        totalqueries = 0
        delta = 0
        for i in range(499):
            current = self._task["ops-temp"][i]
            next = self._task["ops-temp"][i+1]
            totalgets += current["tot-gets"]
            totalsets += current["tot-sets"]
            totalqueries += current["tot-queries"]
            delta += (next["start-time"] - current["start-time"])
        merged["endTime"] = merged["startTime"] + delta
        merged["totalSets"] = totalsets
        merged["totalGets"] = totalgets
        merged["totalQueries"] = totalqueries
        qps = totalqueries / float(delta)
        merged["queriesPerSec"] = qps
        return merged

    def total_stats(self, ops_stat):
        ops_stat["time"] = time.time()
        self._task["totalops"].append(ops_stat)

    def build_stats(self,nodes):
        json_response = StatUtil.build_info(nodes[0])
        self._task["buildstats"] = json_response

    def machine_stats(self,nodes):
        machine_stats = StatUtil.machine_info(nodes[0])
        self._task["machinestats"] = machine_stats

    def reb_stats(self, start, dur):
        print "[reb_stats] recording reb start = {0}, reb duration = {1}"\
            .format(start, dur)
        self._reb_stats["reb_start"] = start
        self._reb_stats["reb_dur"] = dur

    def _extract_proc_info(self, shell, pid):
        o, r = shell.execute_command("cat /proc/{0}/stat".format(pid))
        fields = ('pid comm state ppid pgrp session tty_nr tpgid flags minflt '
                  'cminflt majflt cmajflt utime stime cutime cstime priority '
                  'nice num_threads itrealvalue starttime vsize rss rsslim '
                  'startcode endcode startstack kstkesp kstkeip signal blocked '
                  'sigignore sigcatch wchan nswap cnswap exit_signal '
                  'processor rt_priority policy delayacct_blkio_ticks '
                  'guest_time cguest_time ').split(' ')

        d = dict(zip(fields, o[0].split(' ')))
        return d

    def system_stats(self, nodes, pnames, frequency, verbosity=False):
        shells = []
        for node in nodes:
            try:
                bucket = RestConnection(node).get_buckets()[0].name
                MemcachedClientHelper.direct_client(node, bucket)
                shells.append(RemoteMachineShellConnection(node))
            except:
                pass
        d = {"snapshots": []}
        #        "pname":"x","pid":"y","snapshots":[{"time":time,"value":value}]

        start_time = str(self._task["time"])
        while not self._aborted():
            time.sleep(frequency)
            current_time = time.time()
            i = 0
            for shell in shells:
                node = nodes[i]
                unique_id = node.ip+'-'+start_time
                for pname in pnames:
                    obj = RemoteMachineHelper(shell).is_process_running(pname)
                    if obj and obj.pid:
                        value = self._extract_proc_info(shell, obj.pid)
                        value["name"] = pname
                        value["id"] = obj.pid
                        value["unique_id"] = unique_id
                        value["time"] = current_time
                        value["ip"] = node.ip
                        d["snapshots"].append(value)
                i +=  1
        self._task["systemstats"] = d["snapshots"]
        print " finished system_stats"

    def couchdb_stats(nodes):
        pass

    def capture_mb_snapshot(self, node):
        """
        Capture membase stats snapshot manually
        """
        print "[capture_mb_snapshot] capturing memcache stats snapshot for {0}"\
            .format(node.ip)
        stats = {}

        try:
            bucket = RestConnection(node).get_buckets()[0].name
            mc = MemcachedClientHelper.direct_client(node, bucket)
            stats = mc.stats()
        except Exception as e:
            print "[capture_mb_snapshot] Exception: {0}".format(str(e))
            return False

        stats["time"] = time.time()
        stats["ip"] = node.ip
        self._mb_stats["snapshots"].append(stats)

        print stats
        print "[capture_mb_snapshot] memcache stats snapshot captured"
        return True

    def membase_stats(self,nodes, bucket, frequency, verbose=False):
        mcs = []
        for node in nodes:
            try:
                bucket = RestConnection(node).get_buckets()[0].name
                mcs.append(MemcachedClientHelper.direct_client(node, bucket))
            except:
                pass
        self._task["membasestats"] = []
        self._task["timings"] = []
        self._task["dispatcher"] = []
        d = {}
        #        "pname":"x","pid":"y","snapshots":[{"time":time,"value":value}]
        for mc in mcs:
            d[mc.host] = {"snapshots": [], "timings":[], "dispatcher":[]}

        while not self._aborted():
            time_left = frequency
            timings = None
            # at minimum we want to check for aborted every minute
            while not self._aborted() and time_left > 0:
                time.sleep(min(time_left, 60))
                time_left -= 60
            for mc in mcs:
                retries = 0
                stats = {}
                while not stats and retries < RETRIES:
                    try:
                        stats = mc.stats()
                        try:
                            mem_stats = mc.stats('raw memory')
                        except MemcachedError:
                            mem_stats = mc.stats('memory')
                        stats.update(mem_stats)
                    except Exception as e:
                        print "[memebase_stats] Exception: {0}, retries = {1}"\
                            .format(str(e), retries)
                        time.sleep(2)
                        mc.reconnect()
                        retries += 1
                        continue
                stats["time"] = time.time()
                stats["ip"] = mc.host
                d[mc.host]["snapshots"].append(stats)
                timings = mc.stats('timings')
                d[mc.host]["timings"].append(timings)
                dispatcher = mc.stats('dispatcher')
                d[mc.host]["dispatcher"].append(dispatcher)
            print "\nDumping disk timing stats: {0}".format(time.strftime('%X %x %Z'))
            if timings:
                # TODO dump timings for all servers
                for key, value in sorted(timings.iteritems()):
                    if key.startswith("disk"):
                        print "{0:50s}:     {1}".format(key, value)

        start_time = str(self._task["time"])
        for mc in mcs:
            ip = mc.host
            unique_id = ip+'-'+start_time
            current_time = time.time()
            if self._mb_stats["snapshots"]:
                # use manually captured stats
                self._task["membasestats"] = self._mb_stats["snapshots"]
            else:
                # use periodically captured stats
                for snapshot in d[mc.host]["snapshots"]:
                    snapshot['unique_id'] = unique_id
                    snapshot['time'] = current_time
                    snapshot['ip'] = ip
                    self._task["membasestats"].append(snapshot)
            for timing in d[mc.host]["timings"]:
                timing['unique_id'] = unique_id
                timing['time'] = current_time
                timing['ip'] = ip
                self._task["timings"].append(timing)
            for dispatcher in d[mc.host]["dispatcher"]:
                dispatcher['unique_id'] = unique_id
                dispatcher['time'] = current_time
                dispatcher['ip'] = ip
                self._task["dispatcher"].append(dispatcher)

        print " finished membase_stats"

    def ns_server_stats(self, nodes, bucket, frequency, verbose=False):

        self._task["ns_server_stats"] = []
        self._task["ns_server_stats_system"] = []
        d = {}
        for node in nodes:
            d[node] = {"snapshots": [], "system_snapshots": [] }

        while not self._aborted():
            time.sleep(frequency)
            print "Collecting ns_server_stats"
            for node in nodes:
                rest = RestConnection(node)
                data_json = rest.fetch_bucket_stats(bucket=bucket, zoom='minute')
                d[node]["snapshots"].append(data_json)

                data_json = rest.fetch_system_stats()
                d[node]["system_snapshots"].append(data_json)

        for node in nodes:
           for snapshot in d[node]["snapshots"]:
               self._task["ns_server_stats"].append(snapshot)
           for snapshot in d[node]["system_snapshots"]:
               self._task["ns_server_stats_system"].append(snapshot)

        print " finished ns_server_stats"

    def _aborted(self):
        return self._task["state"] == "stopped"

# Invokes optional callback when registered levels have been reached
# during stats sample()'ing.
#
class CallbackStatsCollector(StatsCollector):

    def __init__(self, verbosity):
        # Tuples of level_name, level, callback.
        self.level_callbacks = []
        super(CallbackStatsCollector, self).__init__(verbosity)

    def sample(self, cur):
        for level_name, level, callback in self.level_callbacks:
            if level < cur.get(level_name, -1):
                callback(cur)

        return super(CallbackStatsCollector, self).sample(cur)

class StatUtil(object):

    @staticmethod
    def build_info(node):
        rest = RestConnection(node)
        api = rest.baseUrl + 'nodes/self'
        status, content, header = rest._http_request(api)
        json_parsed = json.loads(content)
        return json_parsed

    @staticmethod
    def machine_info(node):
        shell = RemoteMachineShellConnection(node)
        info = shell.extract_remote_info()
        return {"type": info.type, "distribution": info.distribution_type,
                "version": info.distribution_version, "ram": info.ram,
                "cpu": info.cpu, "disk": info.disk, "hostname":info.hostname}



#!/usr/bin/env python

class CarbonKey:
    """
    Optionally used by {@class: CarbonMessage}
    """
    def __init__(self, src, server, metric,
                 cluster="", bucket="", vbucket=""):
        self.src = src
        self.server = server
        self.metric = metric
        self.cluster = cluster
        self.bucket = bucket
        self.vbucket = str(vbucket)

    def pack(self):
        buf = self.src + "."
        buf += self.server.replace('.', '-') + "."
        if self.cluster:
            buf += "cluster-" + self.cluster + "."
        if self.bucket:
            buf += "bucket-" + self.bucket + "."
        if self.vbucket:
            buf += "vbucket-" + self.vbucket + "."
        buf += self.metric
        return buf

    def __repr__(self):
        return self.pack()

    def __str__(self):
        return self.pack()



#!/usr/bin/env python

from crc32 import crc32_hash
from multiprocessing.queues import JoinableQueue

try:
    import threading as _threading
except ImportError:
    import dummy_threading as _threading

class VbucketHelper:

    @staticmethod
    def get_vbucket_id(key, num_vbuckets):
        vbucketId = 0
        if num_vbuckets > 0:
            vbucketId = crc32_hash(key) & (num_vbuckets - 1)
        return vbucketId

class UnblockingJoinableQueue(JoinableQueue):
    """
    A joinable queue with methods for unblocking events.
    """

    def all_finished(self):
        """
        Check if all tasks have finished
        """
        self._cond.acquire()
        try:
            if self._unfinished_tasks._semlock._is_zero():
                return True
        finally:
            self._cond.release()

        return False

class SyncDict:
    """
    Synchronized dict wrapper with basic atomic operations.
    """
    def __init__(self):
        self.dict = {}
        self.mutex = _threading.RLock()
        self.not_empty = _threading.Condition(self.mutex)

    def __repr__(self):
        return "<%s> dict: %s" % (self.__class__.__name__, self.dict)

    def put(self, key, value):
        with self.mutex:
            self.dict[key] = value
            self.not_empty.notify()

    def get(self, key):
        with self.mutex:
            try:
                return self.dict[key]
            except KeyError as e:
                print "<%s> failed to get key %s : %s"\
                    % (self.__class__.__name__, key, e)
        return None

    def pop(self, key):
        with self.mutex:
            try:
                return self.dict.pop(key)
            except KeyError as e:
                print "<%s> failed to pop key %s : %s"\
                    % (self.__class__.__name__, key, e)
        return None

    def remove(self, key):
        with self.mutex:
            try:
                del self.dict[key]
            except KeyError as e:
                print "<%s> failed to delete key %s : %s"\
                    % (self.__class__.__name__, key, e)
                return False
        return True

    def wait_not_empty(self):
        """
        Wait until an item becomes available.

        @caution: Not a traditional pro/con call.
                  To minimize locking scope,
                  items won't be consumed here.
        """
        with self.mutex:
            while not self._size():
                self.not_empty.wait()

    def clear(self):
        with self.mutex:
            self.dict.clear()

    def empty(self):
        with self.mutex:
            return not self._size()

    def size(self):
        with self.mutex:
            return self._size()

    def _size(self):
        return len(self.dict)


#!/usr/bin/env python

class DataHelper:
    """
    Helper to retrieve data from particular nodes
    """
    @staticmethod
    def get_bucket(root, parents):
        """Get bucket name from root and parent nodes"""
        path = root.get("path", "")
        if path:
            # assume path like /pools/default
            return path.split("/")[-1]
        else:
            return "default"

    @staticmethod
    def get_ip(root, parent):
        """
        Get hostname or ip address from root and parent nodes

        Carbon needs to know the originator of the fast changing data, \
        for the purpose of contruct the metric info.
        """
        return root.get("host", "127.0.0.1")


import urlparse
import uuid
import time

import logger
import couchdb


log = logger.Logger.get_logger()
class ResultsHelper(object):
    def __init__(self, server):
        parsed = urlparse.urlparse(server,"http")
        self.scheme = parsed.scheme
        self.username = parsed.username
        self.password = parsed.password
        self.hostname = parsed.hostname
        if parsed.port:
            self.port = parsed.port
        else:
            self.port = "5984"
        self.dbname = parsed.path[1:]

        couch = couchdb.Server("http://{0}:{1}@{2}:{3}/".format(self.username, self.password, self.hostname, self.port))
        if not self.dbname in couch:
            couch.create(self.dbname)
            # TODO create views
        self.database = couch[self.dbname]


    def add_perf(self, data, log=None):
        if not "timestamp" in data:
            data["timestamp"] = time.time()
        if not "product" in data:
            data["product"] = "couchbase"
        data["type"] = "performance"
        docid, revid = self.database.save(data)
        if log:
            self.database.put_attachment(docid, log, filename="log", content_type="text/plain")
        return docid


    def compare_perf(self, data, limit, same_keys, different_keys, result_key):
        # returns a list with the percent different for key (ex: "average_ops")
        # between the passed in data and stored data
        # data -- new data that we are testing
        # limit -- limit number of previous results to compare
        # same_keys -- list of keys that must match between new and old
        # different_keys -- dictionary of keys:values that the old data must match
        # result_key -- key that will be compared between new and old

        constraints = "doc.type == 'performance'"
        for key in same_keys:
            if constraints:
                constraints += " && "
            value = data[key]
            if isinstance(value, bool):
                if value == True:
                    constraint = "doc."+key+" = true"
                else:
                    constraint = "doc."+key+" != true"
            elif isinstance(value, int) or isinstance(value, float):
                constraint = "doc."+key+" == " + str(value)
            elif isinstance(value, str):
                constraint = "doc."+key+" == '" + value + "'"
            elif not value:
                constraint = "doc."+key+" == null"
            else:
                constraint = "doc."+key+" == '" + str(value) + "'"
            constraints += constraint

        for key,value in different_keys.iteritems():
            if constraints:
                constraints += " && "
            if isinstance(value, bool):
                if value == True:
                    constraint = "doc."+key+" = true"
                else:
                    constraint = "doc."+key+" != true"
            elif isinstance(value, int) or isinstance(value, float):
                constraint = "doc."+key+" == " + str(value)
            elif isinstance(value, str):
                constraint = "doc."+key+" == '" + value + "'"
            elif not value:
                constraint = "doc."+key+" == null"
            else:
                constraint = "doc."+key+" == '" + str(value) + "'"
            constraints += constraint

        map_fun = '''function(doc) {
  if ('''+constraints+''') {
    emit(doc.timestamp, doc.'''+result_key+''');
  }
}'''
        log.info(map_fun)
        percent_change = []
        if limit:
            rows = self.database.query(map_fun, limit=limit, descending=True)
        else:
            rows = self.database.query(map_fun, descending=True)
        for row in rows:
            percent_change.append(float(data[result_key])/float(row.value))
            log.info("{0} (new / old): {1} / {2} -> {3}".format(result_key, data[result_key], row.value, float(data[result_key])/float(row.value)))

        return percent_change



#!/usr/bin/env python
"""
- Paint the table on:

    - blessings terminal (http://pypi.python.org/pypi/blessings)

- Callbacks for tabula package to format a value based on its meta.
"""

import logging
import ast

from blessings import Terminal
from tabula.table import Table

UNITS_SUFFIX = ["", "K", "M", "G", "T"]

term = Terminal()

STYLES = {"bold": term.bold,
          "reverse": term.reverse,
          "underline": term.underline,
          "blink": term.blink,
          "italic": term.italic,
          "shadow": term.shadow,
          "standout": term.standout,
          "subscript": term.subscript,
          "superscript": term.superscript,
          "flash": term.flash}

COLORS = {"red": term.red,
          "black": term.black,
          "green": term.green,
          "yellow": term.yellow,
          "blue": term.blue,
          "magenta": term.magenta,
          "white": term.white}

def conv_units(val, meta):
    """
    Callback function for tabula

    Format and convert units to be more human readable

    @return new val with converted units
    """
    if not val or not meta:
        return val

    try:
        val = float(val)
    except ValueError:
        logging.error("unable to apply convert units for %s" % val)
        return val

    suf = 0
    if meta in ["bytes", "items"]:

        while val > 1024 and suf < 4:
            val /= 1024
            suf += 1
        return "%.2f%s" % (val, UNITS_SUFFIX[suf])

    return "%.2f" % val

def change_style(val, meta):
    """
    Callback function for tabula

    Change style to display the value, for blessings terminal only

    Supported style: bold, reverse, underline, blink, normal, etc
    """
    if not val or not meta:
        return val

    if meta in STYLES:
        return STYLES[meta](val)

    return val

def change_color(val, meta):
    """
    Callback function for tabula

    Change color to display the value, for blessings terminal only
    """
    if not val or not meta:
        return val

    if meta in COLORS:
        return COLORS[meta](val)

    return val

def check_range(val, meta):
    """
    Callback function for tabula

    Check the range of acceptance defined in metadata
    and display error on screen

    Range is a str literal which evaluates to a @tuple or @list pair
    e.g - (10,100), [20,50]

    Value and edges are converted to float before comparison
    """
    if not val or not meta:
        return val

    try:
        f_val = float(val)
        f_min, f_max = map(lambda x: float(x), ast.literal_eval(meta))
    except (SyntaxError, ValueError), e:
        logging.error("unable to parse range string - %s: %s" % (meta, e))
        return val

    if f_val < f_min or f_val > f_max:
        return COLORS["red"](val)

    return val

"""
Functions to format a value using its meta,
which are called in alphabecial order.

TABULA_CONV_FUNCS: convert value before column alignments
TABULA_DECO_FUNCS: decorate value after column alignments
"""
TABULA_CONV_FUNCS = {"units": conv_units}

TABULA_DECO_FUNCS = {"style": change_style,
                     "color": change_color,
                     "range": check_range}

def enter_fullscreen():
    """
    Invoke before printing out anything.
    This method should be replaced by or merged to blessings package
    """
    term.stream.write(term.enter_fullscreen)
    term.stream.write(term.hide_cursor)

def exit_fullscreen():
    """
    Invoke before printing out anything.
    This method should be replaced by or merged to blessings package
    """
    term.stream.write(term.exit_fullscreen)
    term.stream.write(term.normal_cursor)

def paint(tbl):
    """
    Paint the table on terminal
    """
    if not isinstance(tbl, Table):
        logging.error("unable to paint table: invalid object")
        return False

    term.stream.write(term.clear)

    term.stream.write(str(tbl))
    return True


#!/usr/bin/env python

import time

from obs_def import ObserveKeyState, ObserveStatus
from obs_helper import SyncDict

class Observable:

    def __init__(self, key="", cas=0x0000000000000000,
                 start_time=0, persist_count=1, repl_count=1):
        self.key = key
        self.cas = cas
        self.status = ObserveStatus.OBS_UNKNOWN
        if start_time:
            self.start_time = start_time
        else:
            self.start_time = time.time()
        self.persist_end_time = 0
        self.persist_count = int(persist_count)
        self.persist_servers = set()
        self.repl_end_time = 0
        self.repl_count = int(repl_count)
        self.repl_servers = set()

    def remove_server(self, server, repl=False, stats=True):
        """ remove server from checklist, decrement counter
            and record finish time if necessary
        """
        if repl:
            if server in self.repl_servers:
                self.repl_servers.remove(server)
                self.repl_count -= 1
                if stats and not self.repl_count:
                    self.repl_end_time = time.time()
        else:
            if server in self.persist_servers:
                self.persist_servers.remove(server)
                self.persist_count -= 1
                if stats and not self.persist_count:
                    self.persist_end_time = time.time()

    def __repr__(self):
        return "<%s> key: %s, cas: %x, status: %x, start_time: %d, "\
               "persist_end_time: %d ,repl_end_time: %d, "\
               "persist_count = %d, persist_servers = %s, "\
               "repl_count: %d, repl_servers: %s\n" %\
               (self.__class__.__name__, self.key, self.cas, self.status,
                self.start_time, self.persist_end_time,self.repl_end_time,
                self.persist_count, self.persist_servers,
                self.repl_count, self.repl_servers)

    def __str__(self):
        return self.__repr__()

class Observer:
    #TODO: logging

    # {key_str: Observable}
    _observables = SyncDict()

    def _observe_blocking(self):
        self._observables.wait_not_empty()

        print "<%s> self._observables %s" % (self.__class__.__name__, self._observables)

        if not self._send():
            return False

        responses = self._recv()
        if responses:
            self.update_observables(responses)

    def observe(self, blocking=True):
        if blocking:
            self._observe_blocking()
        else:
            raise NotImplementedError("<%s> unblocking observe has not been implemented"
                % self.__class__.__name__ )

    def load_observables(self, observables):
        """
        Load observables into cache for observation.
        @param observables must be a iterable collections of Observable
        """
        if not observables:
            print "<%s> invalid argument observables: %s" \
                    % (self.__class__.__name__, observables)
            return

        if self._observables.empty():
            print "<%s> load observables = %s" % (self.__class__.__name__, observables)
            for obs in observables:
                self._observables.put(obs.key, obs)

    def clear_observables(self):
        print "clear observables : %s" % self._observables
        self._observables.clear()

    def update_observables(self, responses):
        """
        Update observables based on (@param: responses),
        using key_state as filter
        """
        if not responses:
            print "<%s> empty responses" % self.__class__.__name__
            return True

        if self._observables.empty():
            return True

        res_keys = self.reskey_generator(responses)

        for server, res_key in res_keys:
            obs = self._observables.get(res_key.key)

            if not obs:
                continue

            if server in obs.persist_servers:
                if obs.cas == res_key.cas:
                    if res_key.key_state == ObserveKeyState.OBS_PERSISITED:
                        obs.remove_server(server, repl=False)
                else:
                    obs.status = ObserveStatus.OBS_MODIFIED
            elif server in obs.repl_servers:
                if res_key.key_state in \
                    [ObserveKeyState.OBS_PERSISITED, ObserveKeyState.OBS_FOUND]:
                        obs.remove_server(server, repl=True,
                                          stats=(obs.cas == res_key.cas))
                elif res_key.key_state == ObserveKeyState.OBS_IMPOSSIBLE:
                    print "<%s> invalid key_state %x from repl sever" \
                        % (self.__class__.__name__, server)
                    obs.remove_server(server, repl=True, stats=False)
            else:
                print "<%s> invalid server: %s" \
                    % (self.__class__.__name__, server)
                obs.status = ObserveStatus.OBS_ERROR

            if obs.persist_count <= 0 and \
                obs.repl_count <= 0:
                obs.status = ObserveStatus.OBS_SUCCESS

            self._observables.put(obs.key, obs)
        return True

    def num_observables(self):
        return self._observables.size()

    def reskey_generator(self, responses):
        for server, reses in responses.iteritems():
            for res in reses:
                if res.__class__.__name__ == "ObserveResponse":
                    for res_key in res.keys:
                        yield (server, res_key)

    def observable_filter(self, status):
        """
        Generate observables, using status as filter
        not thread safe.
        """
        for obs in self._observables.dict.itervalues():
            if obs.status == status:
                yield obs

    def _build_conns(self):
        raise NotImplementedError(
            "<%s> _build_conns() has not been implemented" % self.__class__.__name__)

    def _send(self):
        """send packets to servers"""
        raise NotImplementedError(
            "<%s> _send() has not been implemented" % self.__class__.__name__)

    def _recv(self):
        """receive packets from servers"""
        raise NotImplementedError(
            "<%s> _recv() has not been implemented" % self.__class__.__name__)

